#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Lloyd Carothers
Qt Wizard for selecting responses from the IRIS NRL
"""

from typing import Union
from PySide6.QtWidgets import (
    QWizard,
    QWizardPage,
    QVBoxLayout,
    QScrollArea,
    QFrame,
    QLabel,
    QLineEdit,
    QRadioButton
    )
from obspy.clients.nrl import NRL

SENSOR = object()
DATALOGGER = object()

SR = 'sr'
GAIN = 'gain'


class NRLWizard(QWizard):
    """
    A wizard for selecting responses from the IRIS NRL.

    This wizard guides users through the process of selecting either sensor
    or datalogger responses from the IRIS Network Response Library.

    Attributes:
        nrl: NRL client instance
        top: Top level of NRL hierarchy
        q_and_as: List of questions and answers
    """
    def __init__(self, equipment: object = SENSOR, root: str = None) -> None:
        """
        Initialize the NRL Wizard.

        Args:
            equipment: Type of equipment to select (SENSOR or DATALOGGER)
            root: Root directory for NRL data. If None, uses default NRL location.

        Raises:
            ValueError: If equipment type is invalid.
        """
        super().__init__()
        # Hack for now as NRL(root=None) crashes
        if not root:
            self.nrl = NRL()
        else:
            self.nrl = NRL(root=root)
        nrl_root = self.nrl.root

        if equipment is SENSOR:
            self.top = self.nrl.sensors
            self.setWindowTitle(f'Select a sensor from NRL {nrl_root}')
        elif equipment is DATALOGGER:
            self.top = self.nrl.dataloggers
            self.setWindowTitle(f'Select a datalogger from NRL {nrl_root}')
        else:
            raise ValueError("Invalid equipment type")
        self.q_and_as = []
        self.addPage(NRLPage(self, self.top))

    def fresh_page(self):
        """
        Remove all pages ahead of the current page.
        Used when navigating back in the wizard.
        """
        # remove all pages ahead
        current_id = self.currentId()
        for id_ in self.pageIds():
            if id_ > current_id:
                self.removePage(id_)
        # build next page
        question = self.top
        for answer in (a for q, a in self.q_and_as):
            question = question[answer]
        self.addPage(NRLPage(self, question))


class NRLPage(QWizardPage):
    """
    A wizard page for NRL response selection.

    Displays either a list of radio buttons for response selection
    or a final page with response details and nickname field.
    """
    def __init__(self, parent: NRLWizard, nrl_dict: Union[dict, tuple]) -> None:
        """
        Initialize NRL page.

        Args:
            parent: Parent NRLWizard instance
            nrl_dict: Dictionary of NRL options or final response tuple
        """
        super().__init__(parent)
        self.parent_ = parent
        self.nrl_dict = nrl_dict
        self.final = False
        self.selected_radio = None
        self.nickname = ''

    def initializePage(self) -> None:
        """Initialize the page layout and widgets."""
        self.setLayout(QVBoxLayout())

        scroll_area = QScrollArea(self)
        self.layout().addWidget(scroll_area)
        scroll_area.setWidgetResizable(True)

        inner = QFrame(scroll_area)
        layout = QVBoxLayout(inner)
        inner.setLayout(layout)
        scroll_area.setWidget(inner)

        # Last page
        if isinstance(self.nrl_dict, tuple):
            self.final = self.nrl_dict
            layout.addWidget(QLabel(str(self.nrl_dict)))
            layout.addWidget(QLabel('Name:'))
            self.nickname = QLineEdit()
            layout.addWidget(self.nickname)
            self.nickname.setText(self.nrl_dict[0])
        # q and a
        else:
            question = self.nrl_dict._question
            q = QLabel(question + ':', self)
            layout.addWidget(q)
            for answer in sorted(self. nrl_dict.keys()):
                radio = QRadioButton(answer)
                radio.clicked.connect(self.on_selected)
                radio.q_and_a = (question, answer)
                layout.addWidget(radio)

            # Make a dummy page for next button instead of Done.
            self.parent_.addPage(NRLPage(self.parent_, None))

    def on_selected(self) -> None:
        self.selected_radio = self.sender()

    def validatePage(self) -> int:
        if self.final:
            self.parent_.final = self.final
            self.parent_.nickname = self.nickname.text()
            return 1
        if self.selected_radio:
            self.parent_.q_and_as.append(self.selected_radio.q_and_a)
            self.parent_.fresh_page()
            return 1
        return 0

    def cleanupPage(self) -> None:
        try:
            self.parent_.q_and_as.pop()
        except IndexError:
            pass
