#!/usr/bin/env python
# -*- coding: utf-8 -*-*#
"""
Dialog for selecting stations to copy
Datalogger Type, Sensor Type and Gain
"""

import os
from typing import Any

from PySide6.QtGui import QFontMetrics, Qt
from PySide6.QtWidgets import QTableWidgetItem, QAbstractItemView, QHeaderView
from PySide6.QtUiTools import loadUiType

from nexus.obspy_improved import utc_to_str


def load_ui(filename: str) -> object:
    """
    Helper function.
    Loads a ui file relative to this source file.

    Args:
        filename (str): Name of ui file to load.

    Raises:
        e: Exception encountered creating path
           from given filename.

    Returns:
        object: Returns Qt object from ui file.
    """
    path = os.path.join(os.path.dirname(__file__), filename)
    try:
        ret = loadUiType(path)
    except Exception as e:
        print(e)
        raise e
    return ret


class StationSelectDialog(*load_ui('StationSelectDialog.ui')):
    """
    Dialog for selecting stations to copy.

    Attributes:
        stations (List[Any]): List of stations.
        sta_map (Dict[int, Any]): Mapping of row indices to stations.
        action (str): Action to perform.
        values (Any): Additional values for the action.
    """
    def __init__(self, stations: list[Any], action: str,
                 values: list[str], parent=None):
        """
        Initialize the StationSelectDialog.

        Args:
            stations (List[Any]): List of stations.
            action (str): Action to perform.
            values (Any): Additional values for the action.
            parent (Optional[QDialog]): Parent widget. Defaults to None.
        """
        super().__init__(parent)
        self.stations = stations
        self.sta_map = {}
        self.action = action
        self.values = values
        self.selected_stations = []
        self.setupUi()

    def accept(self) -> None:
        """
        Accept the dialog and store the selected stations.
        """
        rows = set(item.row() for item in self.tableWidget.selectedItems())
        for row in rows:
            self.selected_stations.append(self.sta_map[row])
        super().accept()

    def sort_table(self, column: int) -> None:
        """
        Sorts the station selection table based on column
        header selected by user.

        Args:
            column (int): The index of the user selected
                          column.
        """
        row_map = {0: 'code', 1: 'start_date', 2: 'end_date'}
        order = self.tableWidget.horizontalHeader().sortIndicatorOrder()
        descending = order != Qt.AscendingOrder
        self.sta_map = {ind: v for ind, (_, v) in enumerate(
            sorted(
                self.sta_map.items(),
                key=lambda item: getattr(item[1], row_map[column]),
                reverse=descending
            )
        )}

    def setupUi(self) -> None:
        super().setupUi(self)
        self.setWindowTitle('Select Stations')
        if self.action == 'type & gain':
            self.uiGainCB.setChecked(True)
            self.uiGainCB.setText(
                f"Copy Datalogger Gain: {self.values[0]}"
            )
            self.uiTypeCB.setChecked(True)
            self.uiTypeCB.setText(
                f"Copy Datalogger Type: {self.values[1]}"
            )
            self.uiSensorLabel.setParent(None)
        else:
            self.uiSensorLabel.setText(
                f"Copy Sensor Type: {self.values[0]}"
            )
            self.uiTypeCB.setParent(None)
            self.uiGainCB.setParent(None)
        self.tableWidget.setSelectionMode(
            QAbstractItemView.ExtendedSelection
        )
        self.tableWidget.setRowCount(len(self.stations))
        self.tableWidget.horizontalHeader().sectionClicked.connect(
            self.sort_table
        )
        # Add stations to table
        for row, sta in enumerate(self.stations):
            for col, value in enumerate((f"{sta.location_code:>2}",
                                         utc_to_str(sta.start_date),
                                         utc_to_str(sta.end_date))):
                item = QTableWidgetItem()
                item.setText(value)
                # shorten start/end time
                if col in (1, 2):
                    # full time as tooltip
                    item.setToolTip(value)
                    # rounded time in table
                    value = QFontMetrics(self.tableWidget.font()).elidedText(
                        value,
                        Qt.ElideRight,
                        100
                    )
                    item.setText(value)
                self.tableWidget.setItem(row, col, item)
            self.sta_map[row] = sta
        self.tableWidget.resizeColumnsToContents()
        self.tableWidget.horizontalHeader().setSectionResizeMode(
            QHeaderView.Fixed
        )
